# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from .aves import get_semi_aves
from .cifar import get_cifar
from .eurosat import get_eurosat
from .imagenet import get_imagenet
from .medmnist import get_medmnist
from .stl10 import get_stl10
from .svhn import get_svhn
